import types
from Products import PloneMallRegistry
from AccessControl import allow_class
class InvalidFilter(Exception): pass

def _flatten(tuple):
    res = ()
    for item in tuple:
        if type(item) in (types.TupleType, types.ListType):
            res = res + _flatten(item)
        elif item is not None:
            res = res + (item,)
    return res

def getAllInterfacesFrom(klass):
    try:
        res = list(klass.__implements__)
    except:
        res = []
    for b in klass.__bases__:
        res.extend(list(getAllInterfacesFrom(b)))
    return res

class MallFactory:
    '''
    PloneMall Factory
    '''

    def __init__(self,id,class_definition,interface_definition,**kw):
        self.id = id
        self.class_definition = class_definition
        self.interface_definition = interface_definition
        self.__dict__.update(kw)
        
    def getName(self):
        '''
        Return factory name
        '''

        return self.id

    def __call__(self, *args, **kwargs):
        return self.class_definition(*args, **kwargs)

    def getClass(self):
        '''
        Return class name
        '''
        
        return self.class_definition
    
    def getInterface(self):
        '''
        Return interface name
        '''

        return self.interface_definition

    def constructInstance(self, container, id, ignore_doubles=1, *a, **kw):
        o = self(id)
        container._setObject(id, o)
        o = getattr(container, id) 
        o.initializeArchetype()
        o.update(**kw)      
        return o
    
    def _getFactoryMethod(self, container, check_security=1):
        if not self.product or not self.factory:
            raise ValueError, ('Product factory for %s was undefined' %
                               self.getId())
        p = container.manage_addProduct[self.product]
        m = getattr(p, self.factory, None)
        if m is None:
            raise ValueError, ('Product factory for %s was invalid' %
                               self.getId())
        if not check_security:
            return m
        if getSecurityManager().validate(p, p, self.factory, m):
            return m
        raise AccessControl_Unauthorized( 'Cannot create %s' % self.getId() )


allow_class(MallFactory)

class MallRegistry:
    '''
    A factory registry for PloneMall components
    '''

    def __init__(self):
        self._mr = {}
        # mapping of interface names to interface definitions
        self._registered_interfaces={} 

    def registerMallFactory(self, id, class_definition=None, interface_definition = None, **kw):
        '''
        Register a PloneMall factory to make possible the integration
        between the different PloneMall Products.

        @param id An unique identifier (string).
        @param class_definition An archetypes class definition.
        @param interface_definition PloneMall interface definitions.
        '''

        # TODO: Phil, Why this checking was turned off?
        # Phil's Answer to TODO: because it breaks when you do a refresh or a reinstall

        #if self._mr.has_key(id):
        #    raise InvalidFilter("factory with %s unique_name already exists" % id)
        
        if id and not class_definition:
            class_definition=id
            id=class_definition.__name__

        interfaces=[i for i in _flatten(getAllInterfacesFrom(class_definition)) if hasattr(PloneMallRegistry, i.__name__)]

        self._mr[id] = MallFactory(id, class_definition, interfaces, **kw)

        for iface in interfaces:
            self._registered_interfaces[iface.__name__]=iface

    def unregisterMallFactory(self, id):
        '''
        Unregistry a PloneMall factory.

        @param unique_name An unique identifier (string).
        '''

        del self._mr[id]

    def getMallFactory(self, id):
        '''
        Return a registered PloneMall factory.

        @param unique_name An unique identifier (string).
        '''

        return self._mr[id]
    
    def getInterfaceDefinition(self,iface):
        ''' 
        Return the interface class given a certain name

        This method is useful for meta-programming. That's a good way
        to avoid import all interface definitions from
        PloneMallRegistry module manually.
        '''

        if type(iface) in (type(''),type(u'')):
            try:
                return self._registered_interfaces[iface]
            except KeyError:
                return None
        else: 
            return iface

    def getMallFactoriesOf(self, interface_definition):
        '''
        Return a list of factories that implement a PloneMall interface.

        @param interface_definition PloneMall interface definitions.
        '''

        interface_definition=self.getInterfaceDefinition(interface_definition)
        ret = []
        for k, v in self._mr.items():
            c, t = v.class_definition, v.interface_definition
            for i in t:
                if i is interface_definition:
                    ret.append(v)
                    break
        return ret

    def classSupportsInterface(self, klass, interface):
        '''
        Determines if a class supports a certain interface in the registry.
        '''

        factories = self.getMallFactoriesOf(interface)
        return klass in [f.getClass() for f in self.getMallFactoriesOf(interface)]

    def objectSupportsInterface(self, obj, interface):
        '''
        Determines if an object supports a certain interface in the registry.
        '''

        return self.classSupportsInterface(obj.__class__, interface)

    def getMallFactoryIdsOf(self, interface_definition):
        '''
        Return a list of factory names that implements a PloneMall interface.

        @param interface_definition PloneMall interface definition.
        '''

        return [r.id for r in self.getMallFactoriesOf(interface_definition)]

    def getInterfaces(self):
        '''
        Return a list of registered interface names.
        '''

        res = []
        for v in self._mr.values():
            c = v.class_definition, 
            ii = v.interface_definition
            if type(ii) not in (type(()), type([])):
                ii = (ii)
            for i in ii:
                if i not in res:
                    res.append(i)
        return res

    def getInterfaceIds(self):
        '''
        Return a list of interface ids
        '''

        return [i.getName() for i in self.getInterfaces()]
    
FactoryRegistry = MallRegistry()
